"""
Translations contains all translation like MoveGenerator classes.

.. inheritance-diagram:: fullrmc.Generators.Translations
    :parts: 1

+-----------------------------------------------------+-----------------------------------------------------+
|.. figure:: 500randomTranslations.png                |.. figure:: 10TranslationsAlongAxis.png              |
|   :width: 375px                                     |   :width: 375px                                     |
|   :height: 300px                                    |   :height: 300px                                    |
|   :align: left                                      |   :align: left                                      |
|                                                     |                                                     |
|   Random translation vectors generated from         |   Random translation vectors generated from         |
|   atom at origin. (:class:`TranslationGenerator`)   |   atom at origin along a pre-defined axis.          |
|                                                     |   (:class:`TranslationAlongAxisGenerator`)          |
|                                                     |                                                     |
|                                                     |                                                     |
|                                                     |                                                     |
|                                                     |                                                     |
+-----------------------------------------------------+-----------------------------------------------------+
|.. figure:: translationAlongSymmetryAxis.png         | .. figure:: translationTowardsAxis.png              |
|   :width: 375px                                     |    :width: 375px                                    |
|   :height: 300px                                    |    :height: 300px                                   |
|   :align: left                                      |    :align: left                                     |
|                                                     |                                                     |
|   Random translation vector generated along a       |    Random translation vectors generated towards an  |
|   predefined axis or one of the symmetry axes of the|    axis within some maximum angle.                  |
|   hexane molecule and applied on all the molecule's |    Legend is formatted as axis (angle) (direction)  |
|   atoms at the same time.                           |    (:class:`TranslationTowardsAxisGenerator`        |
|   (:class:`TranslationAlongAxisGenerator`           |    :class:`TranslationTowardsSymmetryAxisGenerator`)|
|   :class:`TranslationAlongSymmetryAxisGenerator`)   |                                                     |
|                                                     |                                                     |
+-----------------------------------------------------+-----------------------------------------------------+
|.. figure:: translationTowardsCenter.png             |                                                     |
|   :width: 375px                                     |                                                     |
|   :height: 300px                                    |                                                     |
|   :align: left                                      |                                                     |
|                                                     |                                                     |
|   Random translation vectors generated towards a    |                                                     |
|   pre-defined center or towards the geometric center|                                                     |
|   of a group of atoms. Here 20 vectors are generated|                                                     |
|   within a maximum separation angle of 30 deg.      |                                                     |
|   (:class:`TranslationTowardsCenterGenerator`)      |                                                     |
|                                                     |                                                     |
|                                                     |                                                     |
+-----------------------------------------------------+-----------------------------------------------------+

.. raw:: html

    <iframe width="560" height="315"
    src="https://www.youtube.com/embed/YRTrsDrVSvI?rel=0"
    frameborder="0" allowfullscreen>
    </iframe>

    <p></p>

    <iframe width="560" height="315"
    src="https://www.youtube.com/embed/Ik0RSQT4DzQ?rel=0"
    frameborder="0" allowfullscreen>
    </iframe>

"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, get_path, get_principal_axis, generate_vectors_in_solid_angle, generate_random_float, generate_random_vector
from ..Core.MoveGenerator import MoveGenerator, PathGenerator


class TranslationGenerator(MoveGenerator):
    """
    Generates random translations moves upon groups of atoms.

    :Parameters:
        #. group (None, Group): The group instance.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0.
           If tuple is given, it is the limits of translation boundaries as
           [min,max] where min>=0 and max>min.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random translations.
        # Maximum translation amplitude is set to 0.3A to all defined groups
        for g in ENGINE.groups:
            g.set_move_generator( TranslationGenerator(amplitude=0.3) )


    """
    def __init__(self, group=None, amplitude=0.2):
        super(TranslationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)

    @property
    def amplitude(self):
        """Translation amplitude limits."""
        return self.__amplitude

    def set_amplitude(self, amplitude):
        """
        Set maximum translation vector allowed amplitude.

        :Parameters:
            #. amplitude (number, tuple): The translation amplitude in
               Angstroms.
               If number is given, it is the maximum translation amplitude in
               Angstroms and must be bigger than 0.
               If tuple is given, it is the limits of translation boundaries
               as [min,max] where min>=0 and max>min.
        """
        if isinstance(amplitude, (list, tuple, set)):
            assert len(amplitude) == 2, LOGGER.error("Translation amplitude tuple must have exactly two items")
            assert is_number(amplitude[0]), LOGGER.error("Translation amplitude first item must be a number")
            assert is_number(amplitude[1]), LOGGER.error("Translation amplitude second item must be a number")
            min = FLOAT_TYPE(amplitude[0])
            max = FLOAT_TYPE(amplitude[1])
            assert min>=0, LOGGER.error("Translation amplitude first item must be bigger than 0")
            assert max>min, LOGGER.error("Translation amplitude first item must be bigger than the second item")
            amplitude = (min,max)
        else:
            assert is_number(amplitude), LOGGER.error("Translation amplitude must be a number")
            amplitude = float(amplitude)
            assert amplitude>0, LOGGER.error("Translation amplitude must be bigger than 0")
            amplitude = (FLOAT_TYPE(0),FLOAT_TYPE(amplitude))
        self.__amplitude = amplitude

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        return True, ""

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Any python object. Not used in this
               generator.
        """
        # translate and return
        return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                  maxAmp=self.__amplitude[1])


class TranslationAlongAxisGenerator(TranslationGenerator):
    """
    Generates random translation moves upon groups of atoms along a
    pre-defined axis.

    :Parameters:
        #. group (None, fullrmc.Engine): The constraint stochastic engine.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0.
           If tuple is given, it is the limits of translation boundaries as
           [min,max] where min>=0 and max>min.
        #. axis (list,set,tuple,numpy.ndarray): The pre-defined translation
           axis vector.
        #. direction (None, True, False): Whether to generate translation
           vector in the same direction of axis or not.
           If None is given, generated axis can be in the same direction of
           axis or in  the opposite.
           If True is given, all generated vectors are in the same direction
           of axis.
           If False is given, all generated vectors are in the opposite
           direction of axis.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationAlongAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations along pre-defined axis (1,1,1).
        # Maximum translation amplitude is set to 0.3A to all defined groups
        for g in ENGINE.groups:
            g.set_move_generator( TranslationAlongAxisGenerator(amplitude=0.3, axis=(1,1,1)) )

    """
    def __init__(self, group=None, amplitude=0.2, axis=(1,0,0), direction=None):
        super(TranslationAlongAxisGenerator, self).__init__(group=group, amplitude=amplitude)
        # set axis
        self.set_axis(axis)
        # set direction
        self.set_direction(direction)

    @property
    def axis(self):
        """Translation axis."""
        return self.__axis

    @property
    def direction(self):
        """ Generated translation vectors direction."""
        return self.__direction

    def set_axis(self, axis):
        """
        Set the axis along which the translation will be performed.

        :Parameters:
            #. axis (list,set,tuple,numpy.ndarray): Translation axis vector.
        """
        assert isinstance(axis, (list,set,tuple,np.ndarray)), LOGGER.error("axis must be a list")
        axis = list(axis)
        assert len(axis)==3, LOGGER.error("axis list must have 3 items")
        for pos in axis:
            assert is_number(pos), LOGGER.error( "axis items must be numbers")
        axis = [FLOAT_TYPE(pos) for pos in axis]
        axis =  np.array(axis, dtype=FLOAT_TYPE)
        self.__axis = axis/FLOAT_TYPE( np.linalg.norm(axis) )

    def set_direction(self, direction):
        """
        Sets the generated translation vectors direction.

        :Parameters:
            #. direction (None, True, False): Whether to generate translation
               vector in the same direction of axis or not. If None generated
               axis can be in the same direction of axis or in the opposite.
               If True all generated vectors are in the same direction of axis.
               If False all generated vectors are in the opposite direction
               of axis.
        """
        assert direction in (None, True, False), LOGGER.error("direction can only be None, True or False")
        self.__direction = direction

    def transform_coordinates(self, coordinates, argument=None):
        """
        translates coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Any python object. Not used in this generator.
        """
        # get translation amplitude
        maxAmp = self.amplitude[1]-self.amplitude[0]
        if self.__direction is None:
            amplitude = (1-2*generate_random_float())*maxAmp
        elif self.__direction:
            amplitude = generate_random_float()*maxAmp
        else:
            amplitude = -generate_random_float()*maxAmp
        # compute baseVector
        baseVector = FLOAT_TYPE( np.sign(amplitude)*self.__axis*self.amplitude[0] )
        # compute translation vector
        vector = baseVector + self.__axis*FLOAT_TYPE(amplitude)
        # translate and return
        return coordinates+vector


class TranslationTowardsAxisGenerator(TranslationAlongAxisGenerator):
    """
    Generates random translation moves upon groups of atoms towards a
    pre-defined axis within a tolerance angle between translation vectors
    and the pre-defined axis.

    :Parameters:
        #. group (None, fullrmc.Engine): The constraint stochastic engine.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0. If tuple is given, it is the
           limits of translation boundaries as [min,max] where min>=0 and
           max>min.
        #. axis (list,set,tuple,numpy.ndarray): The pre-defined translation
           axis vector.
        #. angle (number): The maximum tolerance angle in degrees between a
           generated translation vector and the pre-defined axis.
        #. direction (None, True, False): Whether to generate translation
           vector in the same direction of axis or not.
           If None generated axis can be in the same direction of axis or in
           the opposite.
           If True all generated vectors are in the same direction of axis.
           If False all generated vectors are in the opposite direction of axis.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationTowardsAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations towards a pre-defined axis (1,1,1) within 10 degrees.
        # Maximum translation amplitude is set to 0.3A to all defined groups
        for g in ENGINE.groups:
            g.set_move_generator( TranslationTowardsAxisGenerator(amplitude=0.3, axis=(1,1,1), angle=10) )

    """
    def __init__(self, group=None, amplitude=0.2, axis=(1,0,0), angle=30, direction=True):
        super(TranslationTowardsAxisGenerator, self).__init__(group=group, amplitude=amplitude, axis=axis, direction=direction)
        # set angle
        self.set_angle(angle)

    @property
    def angle(self):
        """ Tolerance maximum angle in rad."""
        return self.__angle

    def set_angle(self, angle):
        """
        Set the tolerance maximum angle.

        :Parameters:
            #. angle (number): The maximum tolerance angle in degrees between
               a generated translation vector and the pre-defined axis.
        """
        assert is_number(angle), LOGGER.error("angle must be numbers")
        assert angle>0, LOGGER.error("angle must be positive")
        assert angle<=360, LOGGER.error("angle must be smaller than 360")
        self.__angle = FLOAT_TYPE(angle)*PI/FLOAT_TYPE(180.)

    def transform_coordinates(self, coordinates, argument=None):
        """
        translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Not used here.
        """
        # generate translation axis
        translationAxis = generate_vectors_in_solid_angle(direction=self.axis,
                                                          maxAngle=self.__angle,
                                                          numberOfVectors=1)[0]
        # get translation amplitude
        maxAmp = self.amplitude[1]-self.amplitude[0]
        if self.direction is None:
            amplitude = (1-2*generate_random_float())*maxAmp
        elif self.direction:
            amplitude = generate_random_float()*maxAmp
        else:
            amplitude = -generate_random_float()*maxAmp
        # compute baseVector
        baseVector = FLOAT_TYPE(np.sign(amplitude)*translationAxis*self.amplitude[0])
        # compute translation vector
        vector = baseVector + translationAxis*FLOAT_TYPE(amplitude)
        # translate and return
        return coordinates+vector


class TranslationAlongSymmetryAxisGenerator(TranslationGenerator):
    """
    Generate random translation moves upon groups of atoms along one of
    their symmetry axis. Only groups containing more than 1 atoms allow
    computing symmetry axis.

    :Parameters:
        #. group (None, Group): The group instance.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0. If tuple is given, it is the
           limits of translation boundaries as [min,max] where min>=0 and
           max>min.
        #. axis (integer): Must be 0,1 or 2 for respectively the mains,
           secondary or tertiary symmetry axis.
        #. direction (None, True, False): Whether to generate translation
           vector in the same direction of axis or not. If None generated
           axis can be in the same direction of axis or in the opposite.
           If True all generated vectors are in the same direction of axis.
           If False all generated vectors are in the opposite direction of axis.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationAlongSymmetryAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations along the second symmetry axis of every group.
        # Maximum translation amplitude is set to 0.3A to all defined groups.
        for g in ENGINE.groups:
            if len(g)>1:
                g.set_move_generator( TranslationAlongSymmetryAxisGenerator(amplitude=0.3, axis=(1,1,1), axis=1) )

    """

    def __init__(self, group=None, amplitude=0.2, axis=0, direction=None):
        super(TranslationAlongSymmetryAxisGenerator, self).__init__(group=group, amplitude=amplitude)
        # set amplitude
        self.set_axis(axis)
        # set direction
        self.set_direction(direction)

    @property
    def axis(self):
        """ Translation axis index."""
        return self.__axis

    @property
    def direction(self):
        """ Generated translation vectors direction."""
        return self.__direction

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)<=1:
            return False, "At least two atoms needed in a group to perform translation along symmetry axis."
        else:
            return True, ""

    def set_axis(self, axis):
        """
        Set the symmetry axis index to translate along.

        :Parameters:
            #. axis (integer): Must be 0,1 or 2 for respectively the main,
               secondary or tertiary symmetry axis
        """
        assert is_integer(axis), LOGGER.error("rotation symmetry axis must be an integer")
        axis = INT_TYPE(axis)
        assert axis>=0, LOGGER.error("rotation symmetry axis must be positive.")
        assert axis<=2, LOGGER.error("rotation symmetry axis must be smaller or equal to 2")
        # convert to radian and store amplitude
        self.__axis = axis

    def set_direction(self, direction):
        """
        Sets the generated translation vectors direction.

        :Parameters:
            #. direction (None, True, False): Whether to generate translation
               vector in the same direction of axis or not.
               If None generated axis can be in the same direction of axis or
               in the opposite. If True all generated vectors are in the same
               direction of axis. If False all generated vectors are in the
               opposite direction of axis.
        """
        assert direction in (None, True, False), LOGGER.error("direction can only be None, True or False")
        self.__direction = direction

    def transform_coordinates(self, coordinates, argument=None):
        """
        translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Not used here.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get translation amplitude
            maxAmp = self.amplitude[1]-self.amplitude[0]
            if self.direction is None:
                amplitude = (1-2*generate_random_float())*maxAmp
            elif self.direction:
                amplitude = generate_random_float()*maxAmp
            else:
                amplitude = -generate_random_float()*maxAmp
            # get axis of translation
            _,_,_,_,X,Y,Z =get_principal_axis(coordinates)
            translationAxis = [X,Y,Z][self.__axis]
            # compute baseVector
            baseVector = FLOAT_TYPE(np.sign(amplitude)*translationAxis*self.amplitude[0])
            # compute translation vector
            vector = baseVector + translationAxis*FLOAT_TYPE(amplitude)
            # translate and return
            return coordinates+vector


class TranslationTowardsSymmetryAxisGenerator(TranslationAlongSymmetryAxisGenerator):
    """
    Generates random translation moves upon groups of atoms towards one of
    its symmetry axis within a tolerance angle between translation vectors
    and the axis. Only groups of more than 1 atom are accepted.

    :Parameters:
        #. group (None, fullrmc.Engine): The constraint stochastic engine.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0. If tuple is given, it is the
           limits of translation boundaries as [min,max] where min>=0 and
           max>min.
        #. axis (integer): Must be 0,1 or 2 for respectively the mains,
           secondary or tertiary symmetry axis.
        #. angle (number): The maximum tolerance angle in degrees between a
           generated translation vector and the pre-defined axis.
        #. direction (None, True, False): Whether to generate translation
           vector in the same direction of axis or not.
           If None generated axis can be in the same direction of axis or in
           the opposite. If True all generated vectors are in the same
           direction of axis. If False all generated vectors are in the
           opposite direction of axis.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationTowardsSymmetryAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations towards the first symmetry axis of every group within 15 degrees.
        # Maximum translation amplitude is set to 0.3A to all defined groups.
        for g in ENGINE.groups:
            if len(g)>1:
                g.set_move_generator( TranslationTowardsSymmetryAxisGenerator(amplitude=0.3, axis=0, angle=15) )

    """

    def __init__(self, group=None, amplitude=0.2, axis=0, angle=30, direction=True):
        super(TranslationTowardsSymmetryAxisGenerator, self).__init__(group=group, amplitude=amplitude, axis=axis, direction=direction)
        # set angle
        self.set_angle(angle)

    @property
    def angle(self):
        """ Tolerance maximum angle in rad."""
        return self.__angle

    def set_angle(self, angle):
        """
        Set the tolerance maximum angle.

        :Parameters:
            #. angle (number): The maximum tolerance angle in degrees between
               a generated translation vector and the pre-defined axis.
        """
        assert is_number(angle), LOGGER.error("angle must be numbers")
        assert angle>0, LOGGER.error("angle must be positive")
        assert angle<=360, LOGGER.error("angle must be smaller than 360")
        self.__angle = FLOAT_TYPE(angle)*PI/FLOAT_TYPE(180.)

    def transform_coordinates(self, coordinates, argument=None):
        """
        translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Not used here.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get axis
            _,_,_,_,X,Y,Z =get_principal_axis(coordinates)
            axis = [X,Y,Z][self.axis]
            # generate translation axis
            translationAxis = generate_vectors_in_solid_angle(direction=axis,
                                                              maxAngle=self.__angle,
                                                              numberOfVectors=1)[0]
            # get translation amplitude
            maxAmp = self.amplitude[1]-self.amplitude[0]
            if self.direction is None:
                amplitude = (1-2*generate_random_float())*maxAmp
            elif self.direction:
                amplitude = generate_random_float()*maxAmp
            else:
                amplitude = -generate_random_float()*maxAmp
            # compute baseVector
            baseVector = FLOAT_TYPE(np.sign(amplitude)*translationAxis*self.amplitude[0])
            # compute translation vector
            vector = baseVector + translationAxis*FLOAT_TYPE(amplitude)
            # translate and return
            return coordinates+vector


class TranslationAlongSymmetryAxisPath(PathGenerator):
    """
    Generates translation moves upon groups of atoms along one of their
    symmetry axis. Only groups of more than 1 atom are accepted.

    :Parameters:
        #. group (None, Group): The group instance.
        #. axis (integer): Must be 0,1 or 2 for respectively the main,
           secondary or tertiary symmetry axis
        #. path (List): List of distances.
        #. randomize (boolean): Whether to pull moves randomly from path
           or pull moves in order at every step.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationAlongSymmetryAxisPath

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations of predefined amplitudes along the first symmetry axis of every group.
        amps = [-0.1, 0.075, -0.05, -0.25, 0.01, 0.02, 0.03, 0.1, 0.3]
        for g in ENGINE.groups:
            if len(g)>1:
                g.set_move_generator( TranslationAlongSymmetryAxisPath(axis=0, path=amps) )

    """
    def __init__(self, group=None,  axis=0, path=None, randomize=False):
        # initialize PathGenerator
        PathGenerator.__init__(self, group=group, path=path, randomize=randomize)
        # set axis
        self.set_axis(axis)

    @property
    def axis(self):
        """ Translation axis index."""
        return self.__axis

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): The Group instance.
        """
        if len(group.indexes)<=1:
            return False, "At least two atoms needed in a group to perform translation along symmetry axis."
        else:
            return True, ""

    def set_axis(self, axis):
        """
        Set the symmetry axis index to translate along.

        :Parameters:
            #. axis (integer): Must be 0,1 or 2 for respectively the main,
               secondary or tertiary symmetry axis
        """
        assert is_integer(axis), LOGGER.error("rotation symmetry axis must be an integer.")
        axis = INT_TYPE(axis)
        assert axis>=0, LOGGER.error("rotation symmetry axis must be positive.")
        assert axis<=2, LOGGER.error("rotation symmetry axis must be smaller or equal to 2.")
        # convert to radian and store amplitude
        self.__axis = axis

    def check_path(self, path):
        """
        Check the generator's path.

        :Parameters:
            #. path (None, list): The list of moves.
        """
        if not isinstance(path, (list, set, tuple, np.ndarray)):
            return False, "path must be a list"
        path = list(path)
        if not len(path):
            return False, "path can't be empty"
        for distance in path:
            if not is_number(distance):
                return False, "path items must be numbers"
        return True, ""

    def normalize_path(self, path):
        """
        Transforms all path distances to floating numbers.

        :Parameters:
            #. path (list): The list of moves.

        :Returns:
            #. path (list): The list of moves.
        """
        return [FLOAT_TYPE(distance) for distance in path]

    def transform_coordinates(self, coordinates, argument):
        """
        Translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (float): The move distance.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get translation amplitude
            amplitude = FLOAT_TYPE(argument)
            # get vector of translation
            _,_,_,_,X,Y,Z =get_principal_axis(coordinates)
            vector = [X,Y,Z][self.__axis]
            # amplify vector
            vector *= FLOAT_TYPE(amplitude)
            # translate and return
            return coordinates+vector


class TranslationTowardsCenterGenerator(TranslationGenerator):
    """
    Generates random translation moves of every atom of the group along
    its direction vector to the geometric center of the group.

    :Parameters:
        #. group (None, Group): The group instance.
        #. center (dict): The center value dictionary. Must have a single
           key and this can only be 'fixed' or 'indexes'.
           If key is fixed, value must be a list or a numpy.array of a point
           coordinates such as [X,Y,Z].
           If key is indexes, value must be a list or a numpy array of indexes.
        #. amplitude (number, tuple): The translation amplitude in Angstroms.
           If number is given, it is the maximum translation amplitude in
           Angstroms and must be bigger than 0. If tuple is given, it is the
           limits of translation boundaries as [min,max] where min>=0 and
           max>min.
        #. angle (None, number): The maximum tolerance angle in degrees
           between a generated translation vector and the computed direction.
           If None is given, all generated translation vectors will be along
           the direction to center.
        #. direction (None, True, False): Whether to generate translation
           vectors pointing towards the center or not. If None generated axis
           can be randomly generated towards the center or away from the center.
           If True is given, all generated vectors point towards the center.
           If False is given, all generated vectors point away from the center.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Translations import TranslationTowardsCenterGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to translations towards the origin defined as (0,0,0) within 20 degrees.
        # Maximum translation amplitude is set to 0.2A to all defined groups.
        for g in ENGINE.groups:
            g.set_move_generator( TranslationTowardsCenterGenerator(amplitude=0.2, center={"fixed":(0,0,0)}, axis=0, angle=25) )

    """
    def __init__(self, group=None, center={"fixed":(0,0,0)}, amplitude=0.1, angle=30, direction=True):
        # initialize TranslationGenerator
        TranslationGenerator.__init__(self, group=group, amplitude=amplitude)
        # set direction
        self.set_direction(direction)
        # set center
        self.set_center(center)
        # set angle
        self.set_angle(angle)

    @property
    def direction(self):
        """ Direction value."""
        return self.__direction

    @property
    def center(self):
        """ Center value."""
        return self.__center

    @property
    def angle(self):
        """ Angle value in rad."""
        return self.__angle

    def set_direction(self, direction):
        """
        Set the generated translation vectors direction.

        :Parameters:
            #. direction (None, True, False): Whether to generate translation
               vector in the same direction of axis or not.
               If None generated axis can be in the same direction of axis or
               in the opposite. If True is given, all generated vectors are in
               the same direction of axis. If False is given, all generated
               vectors are in the opposite direction of axis.
        """
        assert direction in (None, True, False), LOGGER.error("direction can only be None, True or False")
        self.__direction = direction

    def set_angle(self, angle):
        """
        Sets the tolerance maximum angle.

        :Parameters:
            #. angle (None, number): The maximum tolerance angle in degrees
               between a generated translation vector and the computed
               direction. If None is given, all generated translation vectors
               will be along the direction to center.
        """
        if angle is not None:
            assert is_number(angle), LOGGER.error("angle must be numbers")
            assert angle>=0, LOGGER.error("angle must be positive")
            assert angle<=360, LOGGER.error("angle must be smaller than 360")
            if FLOAT_TYPE(angle) == FLOAT_TYPE(0.0):
                angle = None
            else:
                angle = FLOAT_TYPE(angle)*PI/FLOAT_TYPE(180.)
        self.__angle = angle

    def set_center(self, center):
        """
        Set center value.

        :Parameters:
            #. center (dict): The center value dictionary. Must have a single
               key and this can only be 'fixed' or 'indexes'. If key is fixed,
               value must be a list or a numpy.array of a point coordinates
               such as [X,Y,Z]. If key is indexes, value must be a list or a
               numpy array of indexes.
        """
        assert isinstance(center, dict), LOGGER.error("center must be a dictionary")
        assert len(center) == 1, LOGGER.error("center must have a single key")
        key = list(center)[0]
        val = center[key]
        assert isinstance(val, (list,set,tuple,np.ndarray)), LOGGER.error("center value must be a list")
        if isinstance(val, np.ndarray):
            assert len(val.shape) == 1, LOGGER.error("center value must have a single dimension")
        assert len(val)>0, LOGGER.error("center value must be a non-zero list.")
        for v in val:
            assert is_number(v), LOGGER.error("center value item must be numbers")
        if key == "fixed":
            self.__mustCompute = False
            assert len(val) == 3, LOGGER.error("fixed center must have exactly 3 elements corresponding to X,Y and Z coordinates of the center point.")
            val = np.array([FLOAT_TYPE(v) for v in val], dtype=FLOAT_TYPE)
        elif key == "indexes":
            self.__mustCompute = True
            for v in val:
                assert is_integer(v), LOGGER.error("indexes center items be integers")
            val =  np.array([INT_TYPE(v) for v in val], dtype=INT_TYPE)
            for v in val:
                assert v>=0, LOGGER.error("indexes center items be positive integers")
        else:
            self.__mustCompute = None
            raise Exception(LOGGER.error("center key must be either 'fixed' or 'indexes'"))
        # set center
        self.__center = {key:val}

    def __get_amplitude(self):
        # get translation amplitude
        maxAmp = self.amplitude[1]-self.amplitude[0]
        if self.__direction is None:
            amplitude = (1-2*generate_random_float())*maxAmp
        elif self.__direction:
            amplitude = generate_random_float()*maxAmp
        else:
            amplitude = -generate_random_float()*maxAmp
        return amplitude

    def __get_translation_axis(self, direction):
        # normalize direction
        norm = np.sqrt(np.add.reduce(direction**2))
        direction[0] /= norm
        direction[1] /= norm
        direction[2] /= norm
        # get vector
        if self.__angle is None:
            vector = direction
        else:
            vector = generate_vectors_in_solid_angle(direction=direction,
                                                     maxAngle=self.__angle,
                                                     numberOfVectors=1)[0]
        # return
        return vector

    def __get_center(self):
        if self.__mustCompute:
            center = self.group._get_engine().realCoordinates[self.__center["indexes"]]
            center = np.mean(center, axis=0).astype(FLOAT_TYPE)
        else:
            center  = self.__center["fixed"]
        return center

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Not used here.
        """
        # get center
        center = self.__get_center()
        # compute coordinates center
        coordsCenter = np.mean(coordinates, axis=0)
        direction    = center-coordsCenter
        # translation vector
        translationAxis = self.__get_translation_axis(direction)
        # get amplitude
        amplitude = self.__get_amplitude()
        # compute baseVector
        baseVector = FLOAT_TYPE(np.sign(amplitude)*translationAxis*self.amplitude[0])
        # compute translation vector
        vector = baseVector + translationAxis*FLOAT_TYPE(amplitude)
        # translate and return
        return coordinates+vector



# standard libraries imports
import os

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "SiOx.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    ENGINE.set_used_frame('size_distribution/7')
    ENGINE.visualize(boxWidth=0,  representationParams="VDW 0.4 100")
else:
    print mes



"""
PairCorrelationConstraints contains classes for all constraints related to
experimental pair correlation functions.

.. inheritance-diagram:: fullrmc.Constraints.PairCorrelationConstraints
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function
import itertools, os, copy, re

# external libraries imports
import numpy as np
from pdbparser.Utilities.Database import is_element_property, get_element_property
from pdbparser.Utilities.Collection import get_normalized_weighting

# fullrmc imports
from ..Globals import FLOAT_TYPE, LOGGER, PI
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import reset_if_collected_out_of_date, get_caller_frames
from ..Core.Constraint import Constraint, ExperimentalConstraint
from ..Core.pairs_histograms import multiple_pairs_histograms_coords, full_pairs_histograms_coords
from ..Constraints.Collection import ShapeFunction
from ..Constraints.PairDistributionConstraints import PairDistributionConstraint

class PairCorrelationConstraint(PairDistributionConstraint):
    """
    Controls the total pair correlation function (pcf) of the system noted
    as g(r). pcf indicates the probability of finding atomic pairs separated
    by the real space distance r. Theoretically g(r) oscillates about 1.
    Also :math:`g(r) \\rightarrow 1` when :math:`r \\rightarrow \\infty`
    and it takes the exact value of zero for :math:`r` shorter than the
    distance of the closest possible approach of pairs of atoms.\n
    Pair correlation function g(r) and pair distribution function G(r) are
    directly related as in the following:
    :math:`g(r)=1+(\\frac{G(r)}{4 \\pi \\rho_{0} r})`.

    g(r) is calculated after binning all pair atomic distances into a weighted
    histograms of values :math:`n(r)` from which local number densities
    are computed as in the following:

    .. math::
        g(r) = \\sum \\limits_{i,j}^{N} w_{i,j} \\frac{\\rho_{i,j}(r)}{\\rho_{0}}
             = \\sum \\limits_{i,j}^{N} w_{i,j} \\frac{n_{i,j}(r) / v(r)}{N_{i,j} / V}

    Where:\n
    :math:`r` is the distance between two atoms. \n
    :math:`\\rho_{i,j}(r)` is the pair density function of atoms i and j. \n
    :math:`\\rho_{0}` is the  average number density of the system. \n
    :math:`w_{i,j}` is the relative weighting of atom types i and j. \n
    :math:`N` is the total number of atoms. \n
    :math:`V` is the volume of the system. \n
    :math:`n_{i,j}(r)` is the number of atoms i neighbouring j at a distance r. \n
    :math:`v(r)` is the annulus volume at distance r and of thickness dr. \n
    :math:`N_{i,j}` is the total number of atoms i and j in the system. \n

    :Parameters: Refer to :class:`.PairDistributionConstraint`


    +----------------------------------------------------------------------+
    |.. figure:: pair_correlation_constraint_plot_method.png               |
    |   :width: 530px                                                      |
    |   :height: 400px                                                     |
    |   :align: left                                                       |
    +----------------------------------------------------------------------+



    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.PairCorrelationConstraints import PairCorrelationConstraint

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # create and add constraint
        PCC = PairCorrelationConstraint(experimentalData="pcf.dat", weighting="atomicNumber")
        ENGINE.add_constraints(PCC)

    """
    def _update_shape_array(self):
        rmin = self._shapeFuncParams['rmin']
        rmax = self._shapeFuncParams['rmax']
        dr   = self._shapeFuncParams['dr'  ]
        qmin = self._shapeFuncParams['qmin']
        qmax = self._shapeFuncParams['qmax']
        dq   = self._shapeFuncParams['dq'  ]
        if rmax is None:
            if self.engine.isPBC:
                a = self.engine.boundaryConditions.get_a()
                b = self.engine.boundaryConditions.get_b()
                c = self.engine.boundaryConditions.get_c()
                rmax = FLOAT_TYPE( np.max([a,b,c]) + 10 )
            else:
                coordsCenter = np.sum(self.engine.realCoordinates, axis=0)/self.engine.realCoordinates.shape[0]
                coordinates  = self.engine.realCoordinates-coordsCenter
                distances    = np.sqrt( np.sum(coordinates**2, axis=1) )
                maxDistance  = 2.*np.max(distances)
                rmax = FLOAT_TYPE( maxDistance + 10 )
                LOGGER.warn("@%s Better set shape function rmax with infinite boundary conditions. Here value is automatically set to %s"%(self.engine.usedFrame, rmax))
        shapeFunc  = ShapeFunction(engine    = self.engine,
                                   weighting = self.weighting,
                                   qmin=qmin, qmax=qmax, dq=dq,
                                   rmin=rmin, rmax=rmax, dr=dr)
        self._shapeArray = shapeFunc.get_gr_shape_function( self.shellCenters )
        del shapeFunc

    def _reset_standard_error(self):
        # recompute squared deviation
        if self.data is not None:
            totalPCF = self.__get_total_gr(self.data, rho0=self.engine.numberDensity)
            self.set_standard_error(self.compute_standard_error(modelData = totalPCF))

    def update_standard_error(self):
        """ Compute and set constraint's standardError."""
        # set standardError
        totalPDF = self.get_constraint_value()["total"]
        self.set_standard_error(self.compute_standard_error(modelData = totalPDF))


    def __get_total_gr(self, data, rho0):
        """This method is created just to speed up the computation of
        the total gr upon fitting.
        """
        gr = np.zeros(self.histogramSize, dtype=FLOAT_TYPE)
        for pair in self.elementsPairs:
            # get weighting scheme
            wij = self.weightingScheme.get(pair[0]+"-"+pair[1], None)
            if wij is None:
                wij = self.weightingScheme[pair[1]+"-"+pair[0]]
            # get number of atoms per element
            ni = self.engine.numberOfAtomsPerElement[pair[0]]
            nj = self.engine.numberOfAtomsPerElement[pair[1]]
            # get index of element
            idi = self.engine.elements.index(pair[0])
            idj = self.engine.elements.index(pair[1])
            # get Nij
            if idi == idj:
                Nij = ni*(ni-1)/2.0
                Dij = Nij/self.engine.volume
                nij = data["intra"][idi,idj,:]+data["inter"][idi,idj,:]
                gr += wij*nij/Dij
            else:
                Nij = ni*nj
                Dij = Nij/self.engine.volume
                nij = data["intra"][idi,idj,:]+data["intra"][idj,idi,:] + data["inter"][idi,idj,:]+data["inter"][idj,idi,:]
                gr += wij*nij/Dij
        # Multiply by scale factor and deviding by shells volume
        gr /= self.shellVolumes
        # remove shape function
        if self._shapeArray is not None:
            gr -= self._shapeArray
        # Multiply by scale factor
        self._fittedScaleFactor = self.get_adjusted_scale_factor(experimentalData=self.experimentalPDF, modelData=gr, dataWeights=self._usedDataWeights, rho0=rho0)
        if self._fittedScaleFactor != 1:
            Gr = (4.*PI*self.shellCenters*rho0)*(gr-1)
            Gr *= self._fittedScaleFactor
            gr  = 1. +  Gr/(4.*PI*self.shellCenters*rho0)
        # apply multiframe prior and weight
        gr = self._apply_multiframe_prior(gr)
        # convolve total with window function
        if self.windowFunction is not None:
            gr = np.convolve(gr, self.windowFunction, 'same')
        return gr

    def get_adjusted_scale_factor(self, experimentalData, modelData, dataWeights, rho0):
        """Overload to bring back g(r) to G(r) prior to fitting scale factor.
        g(r) -> 1 at high r and this will create a wrong scale factor.
        Overloading can be avoided but it's better to for performance reasons
        """
        #r    = self.shellCenters
        SF = self.scaleFactor
        # check to update scaleFactor
        if self.adjustScaleFactorFrequency:
            if not self.engine.accepted%self.adjustScaleFactorFrequency:
                expGr = (4.*PI*self.shellCenters*rho0)*(experimentalData-1)
                Gr    = (4.*PI*self.shellCenters*rho0)*(modelData-1)
                SF = self.fit_scale_factor(expGr, Gr, dataWeights)
        return SF

    def _on_collector_reset(self):
        pass

    def _get_constraint_value(self, data, applyMultiframePrior=True):
        # http://erice2011.docking.org/upload/Other/Billinge_PDF/03-ReadingMaterial/BillingePDF2011.pdf     page 6
        output = {}
        for pair in self.elementsPairs:
            output["rdf_intra_%s-%s" % pair] = np.zeros(self.histogramSize, dtype=FLOAT_TYPE)
            output["rdf_inter_%s-%s" % pair] = np.zeros(self.histogramSize, dtype=FLOAT_TYPE)
            output["rdf_total_%s-%s" % pair] = np.zeros(self.histogramSize, dtype=FLOAT_TYPE)
        output["total_no_window"] = np.zeros(self.histogramSize, dtype=FLOAT_TYPE)
        for pair in self.elementsPairs:
            # get weighting scheme
            wij = self.weightingScheme.get(pair[0]+"-"+pair[1], None)
            if wij is None:
                wij = self.weightingScheme[pair[1]+"-"+pair[0]]
            # get number of atoms per element
            ni = self.engine.numberOfAtomsPerElement[pair[0]]
            nj = self.engine.numberOfAtomsPerElement[pair[1]]
            # get index of element
            idi = self.engine.elements.index(pair[0])
            idj = self.engine.elements.index(pair[1])
            # get Nij
            if idi == idj:
                Nij = ni*(ni-1)/2.0
                output["rdf_intra_%s-%s" % pair] += data["intra"][idi,idj,:]
                output["rdf_inter_%s-%s" % pair] += data["inter"][idi,idj,:]
            else:
                Nij = ni*nj
                output["rdf_intra_%s-%s" % pair] += data["intra"][idi,idj,:] + data["intra"][idj,idi,:]
                output["rdf_inter_%s-%s" % pair] += data["inter"][idi,idj,:] + data["inter"][idj,idi,:]
            # calculate intensityFactor
            intensityFactor = (self.engine.volume*wij)/(Nij*self.shellVolumes)
            # multiply by intensityFactor
            output["rdf_intra_%s-%s" % pair] *= intensityFactor
            output["rdf_inter_%s-%s" % pair] *= intensityFactor
            output["rdf_total_%s-%s" % pair]  = output["rdf_intra_%s-%s" % pair] + output["rdf_inter_%s-%s" % pair]
            output["total_no_window"]        += output["rdf_total_%s-%s" % pair]
        # remove shape function
        if self._shapeArray is not None:
            output["total_no_window"] -= self._shapeArray
        # multiply by scale factor
        if self.scaleFactor != 1:
            Gr = (4.*PI*self.shellCenters*self.engine.numberDensity)*(output["total_no_window"]-1)
            Gr *= self.scaleFactor
            output["total_no_window"]  = 1. +  Gr/(4.*PI*self.shellCenters*self.engine.numberDensity)
        # apply multiframe prior and weight
        if applyMultiframePrior:
            output["total_no_window"] = self._apply_multiframe_prior(output["total_no_window"])
        # convolve total with window function
        if self.windowFunction is not None:
            output["total"] = np.convolve(output["total_no_window"], self.windowFunction, 'same')
        else:
            output["total"] = output["total_no_window"]
        return output

    def get_constraint_value(self, applyMultiframePrior=True):
        """
        Get constraint's data dictionary value.

        :Parameters:
            #. applyMultiframePrior (boolean): Whether to apply subframe weight
               and prior to the total. This will only have an effect when used
               frame is a subframe and in case subframe weight and prior is
               defined.

        :Returns:
            #. PDFs (dictionary): The PDFs dictionnary, where keys are the
               element wise intra and inter molecular PDFs and values are the
               computed PDFs.
        """
        if self.data is None:
            LOGGER.warn("data must be computed first using 'compute_data' method.")
            return {}
        return self._get_constraint_value(self.data, applyMultiframePrior=applyMultiframePrior)

    @reset_if_collected_out_of_date
    def compute_data(self):
        """ Compute constraint's data."""
        intra,inter = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates,
                                                    basis            = self.engine.basisVectors,
                                                    isPBC            = self.engine.isPBC,
                                                    moleculeIndex    = self.engine.moleculesIndex,
                                                    elementIndex     = self.engine.elementsIndex,
                                                    numberOfElements = self.engine.numberOfElements,
                                                    minDistance      = self.minimumDistance,
                                                    maxDistance      = self.maximumDistance,
                                                    histSize         = self.histogramSize,
                                                    bin              = self.bin,
                                                    ncores           = self.engine._runtime_ncores)
        # update data
        self.set_data({"intra":intra, "inter":inter})
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        totalPCF = self.__get_total_gr(self.data, rho0=self.engine.numberDensity)
        self.set_standard_error(self.compute_standard_error(modelData = totalPCF))
        # set original data
        if self.originalData is None:
            self._set_original_data(self.data)

    def compute_before_move(self, realIndexes, relativeIndexes):
        """
        Compute constraint's data before move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
        """
        intraM,interM = multiple_pairs_histograms_coords( indexes          = relativeIndexes,
                                                          boxCoords        = self.engine.boxCoordinates,
                                                          basis            = self.engine.basisVectors,
                                                          isPBC            = self.engine.isPBC,
                                                          moleculeIndex    = self.engine.moleculesIndex,
                                                          elementIndex     = self.engine.elementsIndex,
                                                          numberOfElements = self.engine.numberOfElements,
                                                          minDistance      = self.minimumDistance,
                                                          maxDistance      = self.maximumDistance,
                                                          histSize         = self.histogramSize,
                                                          bin              = self.bin,
                                                          allAtoms         = True,
                                                          ncores           = self.engine._runtime_ncores)
        intraF,interF = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates[relativeIndexes],
                                                      basis            = self.engine.basisVectors,
                                                      isPBC            = self.engine.isPBC,
                                                      moleculeIndex    = self.engine.moleculesIndex[relativeIndexes],
                                                      elementIndex     = self.engine.elementsIndex[relativeIndexes],
                                                      numberOfElements = self.engine.numberOfElements,
                                                      minDistance      = self.minimumDistance,
                                                      maxDistance      = self.maximumDistance,
                                                      histSize         = self.histogramSize,
                                                      bin              = self.bin,
                                                      ncores           = self.engine._runtime_ncores )
        # set active atoms data
        self.set_active_atoms_data_before_move( {"intra":intraM-intraF, "inter":interM-interF} )
        self.set_active_atoms_data_after_move(None)

    def compute_after_move(self, realIndexes, relativeIndexes, movedBoxCoordinates):
        """
        Compute constraint's data after move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new
               coordinates.
        """
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[relativeIndexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[relativeIndexes] = movedBoxCoordinates
        # calculate pair distribution function
        intraM,interM = multiple_pairs_histograms_coords( indexes          = relativeIndexes,
                                                          boxCoords        = self.engine.boxCoordinates,
                                                          basis            = self.engine.basisVectors,
                                                          isPBC            = self.engine.isPBC,
                                                          moleculeIndex    = self.engine.moleculesIndex,
                                                          elementIndex     = self.engine.elementsIndex,
                                                          numberOfElements = self.engine.numberOfElements,
                                                          minDistance      = self.minimumDistance,
                                                          maxDistance      = self.maximumDistance,
                                                          histSize         = self.histogramSize,
                                                          bin              = self.bin,
                                                          allAtoms         = True,
                                                          ncores           = self.engine._runtime_ncores )
        intraF,interF = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates[relativeIndexes],
                                                      basis            = self.engine.basisVectors,
                                                      isPBC            = self.engine.isPBC,
                                                      moleculeIndex    = self.engine.moleculesIndex[relativeIndexes],
                                                      elementIndex     = self.engine.elementsIndex[relativeIndexes],
                                                      numberOfElements = self.engine.numberOfElements,
                                                      minDistance      = self.minimumDistance,
                                                      maxDistance      = self.maximumDistance,
                                                      histSize         = self.histogramSize,
                                                      bin              = self.bin,
                                                      ncores           = self.engine._runtime_ncores  )
        # set active atoms data
        self.set_active_atoms_data_after_move( {"intra":intraM-intraF, "inter":interM-interF} )
        # reset coordinates
        self.engine.boxCoordinates[relativeIndexes] = boxData
        # compute standardError after move
        dataIntra = self.data["intra"]-self.activeAtomsDataBeforeMove["intra"]+self.activeAtomsDataAfterMove["intra"]
        dataInter = self.data["inter"]-self.activeAtomsDataBeforeMove["inter"]+self.activeAtomsDataAfterMove["inter"]
        totalPCF = self.__get_total_gr({"intra":dataIntra, "inter":dataInter}, rho0=self.engine.numberDensity)
        # set after move standard error
        self.set_after_move_standard_error( self.compute_standard_error(modelData = totalPCF) )

    def compute_as_if_amputated(self, realIndex, relativeIndex):
        """
        Compute and return constraint's data and standard error as if
        given atom is amputated.

        :Parameters:
            #. realIndex (numpy.ndarray): Atom's index as a numpy array
               of a single element.
            #. relativeIndex (numpy.ndarray): Atom's relative index as a
               numpy array of a single element.
        """
        # compute data
        self.compute_before_move(realIndexes=realIndex, relativeIndexes=relativeIndex)
        dataIntra = self.data["intra"]-self.activeAtomsDataBeforeMove["intra"]
        dataInter = self.data["inter"]-self.activeAtomsDataBeforeMove["inter"]
        data      = {"intra":dataIntra, "inter":dataInter}
        # temporarily adjust self.__weightingScheme
        weightingScheme = self.weightingScheme
        relativeIndex = relativeIndex[0]
        selectedElement = self.engine.allElements[relativeIndex]
        self.engine.numberOfAtomsPerElement[selectedElement] -= 1
        WS = get_normalized_weighting(numbers=self.engine.numberOfAtomsPerElement, weights=self._elementsWeight )
        for k in WS:
            WS[k] = FLOAT_TYPE(WS[k])
        self._set_weighting_scheme(WS)
        # compute standard error
        if not self.engine._RT_moveGenerator.allowFittingScaleFactor:
            SF = self.adjustScaleFactorFrequency
            self._set_adjust_scale_factor_frequency(0)
        rho0     = ((self.engine.numberOfAtoms-1)/self.engine.volume).astype(FLOAT_TYPE)
        totalPCF = self.__get_total_gr(data, rho0=rho0)
        standardError = self.compute_standard_error(modelData = totalPCF)
        if not self.engine._RT_moveGenerator.allowFittingScaleFactor:
            self._set_adjust_scale_factor_frequency(SF)
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        # set data
        self.set_amputation_data( {'data':data, 'weightingScheme':self.weightingScheme} )
        # compute standard error
        self.set_amputation_standard_error( standardError )
        # reset weightingScheme and number of atoms per element
        self._set_weighting_scheme(weightingScheme)
        self.engine.numberOfAtomsPerElement[selectedElement] += 1

    def _on_collector_collect_atom(self, realIndex):
        pass

    def _on_collector_release_atom(self, realIndex):
        pass





#



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, FLOAT_TYPE, maxint
from fullrmc.Engine import Engine
from fullrmc.Core.Collection import generate_random_float, get_rotation_matrix, rotate
from fullrmc.Generators.Translations import TranslationGenerator
from fullrmc.Constraints.AtomicCoordinationConstraints import AtomicCoordinationNumberConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
##############################  XY TRANSLATION GENERATOR  ################################
class XYTranslationGenerator(TranslationGenerator):
    def transform_coordinates(self, coordinates, argument=None):
        # generate random vector and ensure it is not zero
        vector    = np.array(1-2*np.random.random(3), dtype=FLOAT_TYPE)
        vector[2] = 0
        norm      = np.linalg.norm(vector)
        if norm == 0:
            while norm == 0:
                vector    = np.array(1-2*np.random.random(3), dtype=FLOAT_TYPE)
                vector[2] = 0
                norm      = np.linalg.norm(vector)
        # normalize vector
        vector /= FLOAT_TYPE( norm )
        # compute baseVector
        baseVector = FLOAT_TYPE(vector*self.amplitude[0])
        # amplify vector
        maxAmp  = FLOAT_TYPE(self.amplitude[1]-self.amplitude[0])
        vector *= FLOAT_TYPE(generate_random_float()*maxAmp)
        vector += baseVector
        # translate and return
        return coordinates+vector


##########################################################################################
#####################################  CREATE ENGINE  ####################################
pdbPath = "system.pdb"
ENGINE = Engine(path=None)
ENGINE.set_pdb(pdbPath)
# add constraints
ACN_CONSTRAINT = AtomicCoordinationNumberConstraint()
ENGINE.add_constraints([ACN_CONSTRAINT])
ACN_CONSTRAINT.set_coordination_number_definition( [ ('Al','Cl',1.5, 2.5, 2, 2),
                                                     ('Al','S', 2.5, 3.0, 2, 2)] )
# add inter-molecular distance constraint
EMD_CONSTRAINT = InterMolecularDistanceConstraint(defaultDistance=1.0)
ENGINE.add_constraints([EMD_CONSTRAINT])
# set TranslationGenerator move generators amplitude
ENGINE.set_groups([[idx] for idx, el in enumerate( ENGINE.allElements ) if el != 'Al'])
[g.set_move_generator(XYTranslationGenerator(g, amplitude=0.1)) for g in ENGINE.groups]


##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def run_normal(nsteps, xyzPath):
    ACN_CONSTRAINT.set_coordination_number_definition( [ ('Al','Cl',1.5, 2.5, 2, 2),
                                                         ('Al','S', 2.5, 3.0, 2, 2)] )
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)


##########################################################################################
#####################################  RUN SIMULATION  ###################################
xyzPath ="trajectory.xyz"
if os.path.isfile(xyzPath): os.remove(xyzPath)
run_normal(10000,    xyzPath)


##########################################################################################
##################################  VISUALIZE SIMULATION  ################################
# create otherParams
otherParams = ["draw material Opaque"]
otherParams.append("draw color red")
otherParams.append( "graphics top cylinder {10 10 9.95} {10 10 10.05} radius 1.5 resolution 100 filled no" )
otherParams.append("draw color blue")
otherParams.append( "graphics top cylinder {10 10 9.95} {10 10 10.05} radius 2.5 resolution 100 filled no" )
otherParams.append("draw color green")
otherParams.append( "graphics top cylinder {10 10 9.95} {10 10 10.05} radius 3.0 resolution 100 filled no" )
otherParams.append("label add Bonds 0/0 0/1")
otherParams.append("label add Bonds 0/0 0/2")
otherParams.append("label add Bonds 0/0 0/3")
otherParams.append("label add Bonds 0/0 0/4")
otherParams.append("label textsize 1.25")
otherParams.append("label textthickness 2")
otherParams.append("color Labels Bonds black")
otherParams.append("axes location off")
otherParams.append("light 0 on")
otherParams.append("light 0 pos {20 20 0}")
otherParams.append("light 1 on")
otherParams.append("light 0 pos {-20 -20 0}")
otherParams.append("light 2 on")
otherParams.append("light 0 pos {-20 20 0}")
otherParams.append("light 3 on")
otherParams.append("light 0 pos {20 -20 0}")
otherParams.append("scale set 0.3")
otherParams.append("display resize 800 800" )

ENGINE.set_pdb(pdbPath)
ENGINE.visualize( commands = ["trajectory.xyz"],
                  boxWidth = 0, bgColor="white",
                  representationParams = 'VDW 0.1 100',
                  otherParams = otherParams)



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, maxint
from fullrmc.Engine import Engine
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
#####################################  CREATE ENGINE  ####################################
pdbPath = "system.pdb"
ENGINE = Engine(path=None)
ENGINE.set_pdb(pdbPath)

# add constraints
B_CONSTRAINT  = BondConstraint()
BA_CONSTRAINT = BondsAngleConstraint()
ENGINE.add_constraints([B_CONSTRAINT, BA_CONSTRAINT])
B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 0.8, 1.1),
                                                                  ('OH2' ,'H2' , 0.8, 1.1)] })
BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 80, 120)] })

# set TranslationGenerator move generators amplitude
[g.moveGenerator.set_amplitude(0.025) for g in ENGINE.groups]


##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def run_normal(nsteps, xyzPath):
    B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 0.8, 1.1),
                                                                      ('OH2' ,'H2' , 0.8, 1.1)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)

def run_distort_1(nsteps, xyzPath):
    B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 0.2, 0.5),
                                                                      ('OH2' ,'H2' , 2.2, 2.5)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)

def run_distort_2(nsteps, xyzPath):
    B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 2.2, 2.5),
                                                                      ('OH2' ,'H2' , 0.2, 0.5)] })
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps*2, xyzFrequency=1, xyzPath=xyzPath, restartPdb=None)


##########################################################################################
#####################################  RUN SIMULATION  ###################################
xyzPath ="trajectory.xyz"
if os.path.isfile(xyzPath): os.remove(xyzPath)
run_normal(2000,    xyzPath)
run_distort_1(2000, xyzPath)
run_distort_2(2000, xyzPath)
run_normal(2000,    xyzPath)


##########################################################################################
##################################  VISUALIZE SIMULATION  ################################
ENGINE.set_pdb(pdbPath)
ENGINE.visualize( commands = ["trajectory.xyz"],
                  boxWidth = 0, bgColor="white",
                  representationParams = 'CPK 1.0 0.2 50 50',
                  otherParams = ["label add Atoms 0/1",
                                 "label add Atoms 0/2",
                                 "label add Bonds 0/0 0/1",
                                 "label add Bonds 0/0 0/2",
                                 "label textsize 1.5",
                                 "label textthickness 2",
                                 "color Labels Atoms black",
                                 "color Labels Bonds black",
                                 "scale set 0.75"] )



# standard libraries imports
from __future__ import print_function
import os
import itertools

# external libraries imports
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

# fullrmc library imports
from fullrmc.Engine import Engine

# create figure
FIG = plt.figure()
FIG.patch.set_facecolor('white')
grid = gridspec.GridSpec(nrows=2, ncols=2)
grid.update(left=0.1, bottom=0.1, right=0.95, wspace=0.05)

totalAx = plt.subplot(grid[0, :])
lowRAx  = plt.subplot(grid[1, 0])
highRAx = plt.subplot(grid[1, 1])


# set axis ticks
highRAx.get_yaxis().set_ticks([])
lowRAx.get_yaxis().set_ticks([])

# Create plotting styles
colors = ["b",'g','r','c','m','y']
markers = ["",'.','+','^','|']
INTRA_STYLES = [r[0] + r[1]for r in itertools.product(['--'], colors)]
INTRA_STYLES = [r[0] + r[1]for r in itertools.product(markers, INTRA_STYLES)]
INTER_STYLES = [r[0] + r[1]for r in itertools.product(['-'], colors)]
INTER_STYLES = [r[0] + r[1]for r in itertools.product(markers, INTER_STYLES)]

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "thf_engine.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    PDF = ENGINE.constraints[0]
else:
    print(mes)


# plot total
output = PDF.get_constraint_value()
totalAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
lowRAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
highRAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
totalAx.plot(PDF.shellCenters, output["pcf"], 'k', linewidth=3.0,  markevery=25, label="total" )
lowRAx.plot(PDF.shellCenters, output["pcf"], 'k', linewidth=3.0,  markevery=25, label="total" )
highRAx.plot(PDF.shellCenters, output["pcf"], 'k', linewidth=3.0,  markevery=25, label="total" )

# plot inter and intra
#intraStyleIndex = 0
#interStyleIndex = 0
#for key in output:
#    val = output[key]
#    if key in ("pdf_total", "pdf"):
#        continue
#    elif "intra" in key:
#        totalAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        lowRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        highRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        intraStyleIndex+=1
#    elif "inter" in key:
#        totalAx.plot(PDF.shellCenters, val, INTER_STYLES[interStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        lowRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        highRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
#        interStyleIndex+=1

# set legend
totalAx.legend(ncol=2, frameon=False, fontsize=20)#, loc=(1.05,-1.25))
totalAx.set_xlim([0,16])
totalAx.set_ylim([-1,6.2])

# remove y ticks labels
lowRAx.set_yticklabels([])
highRAx.set_yticklabels([])

# set x limits
lowRAx.relim()
highRAx.relim()
lowRAx.set_xlim([0.5,3])
lowRAx.set_ylim([-1, 6.2])
highRAx.set_xlim([3,16])
highRAx.set_ylim([0.25, 1.25])

# add title
plt.suptitle("$stdErr=%.6f$"%ENGINE.totalStandardError, fontsize=20)

# show plot
plt.show()



#  ####################################################################################  #
#  ########################### IMPORTING USEFUL DEFINITIONS ###########################  #
## standard library imports
import os, sys

## external libraries imports
import numpy as np
from pdbparser.pdbparser import pdbparser
from pdbparser.Utilities.Modify import set_records_attribute_values

## fullrmc imports
from fullrmc.Engine import Engine
from fullrmc.Core.Group import EmptyGroup
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Generators.Translations import TranslationGenerator
from fullrmc.Generators.Swaps import SwapPositionsGenerator


#  ####################################################################################  #
#  ############################# DECLARE USEFUL VARIABLES #############################  #
# dirname
try:
    DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
except:
    DIR_PATH = ''

# files name
grFileName     = "pdf.exp"
pdbFileName    = "system.pdb"
engineFileName = "system.rmc"
NCORES         = 1
FRESH_START    = False

# engine variables
experimentalDataPath = os.path.join(DIR_PATH, grFileName)
structurePdbPath     = os.path.join(DIR_PATH, pdbFileName)
engineSavePath       = os.path.join(DIR_PATH, engineFileName)



#  ####################################################################################  #
#  ################################### CREATE ENGINE ##################################  #
ENGINE = Engine(path=None)
if not ENGINE.is_engine(engineSavePath) or FRESH_START:
    ENGINE = Engine(path=engineSavePath, freshStart=True)
    ENGINE.set_pdb(structurePdbPath)
    ## create and add pair distribution constraint
    PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=experimentalDataPath, weighting="atomicNumber")
    ENGINE.add_constraints([PDF_CONSTRAINT])
    ## create and add intermolecular distances constraint
    EMD_CONSTRAINT = InterMolecularDistanceConstraint()
    ENGINE.add_constraints([EMD_CONSTRAINT])
    EMD_CONSTRAINT.set_type_definition("element")
    EMD_CONSTRAINT.set_pairs_distance([('Co','Co',2.00),('Co','Mn',2.00),('Co','Ni',2.00),('Co','Li',2.00),('Co','O' ,1.7),
                                       ('Mn','Mn',2.00),('Mn','Li',2.00),('Mn','Ni',2.00),('Mn','O' ,1.7),
                                       ('Ni','Ni',2.00),('Ni','Li',2.00),('Ni','O' ,1.7),
                                       ('Li','Li',2.40),('Li' ,'O',1.8),
                                       ('O' ,'O' ,1.20),])
    ## save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(engineSavePath)
    ## unpack constraints before fitting in case tweaking is needed
    PDF_CONSTRAINT, EMD_CONSTRAINT = ENGINE.constraints

#  ####################################################################################  #
#  ############################### DEFINE DIFFERENT RUNS ##############################  #
def normal_run(numberOfSteps=100000, saveFrequency=10000):
    ## reset groups as atoms
    ENGINE.set_groups_as_atoms()
    ## run engine
    ENGINE.run(numberOfSteps=numberOfSteps, saveFrequency=saveFrequency)

def swaps_run(numberOfSteps=100000, saveFrequency=10000):
    ## reset groups as atoms
    ENGINE.set_groups_as_atoms()
    ## build swap lists
    ALL_ELEMENTS = ENGINE.get_original_data('allElements')
    liSwaps = [[idx] for idx in range(len(ALL_ELEMENTS)) if ALL_ELEMENTS[idx]=='li' or ALL_ELEMENTS[idx]=='Li']
    meSwaps = [[idx] for idx in range(len(ALL_ELEMENTS)) if ALL_ELEMENTS[idx] in ('co','ni','mn') or ALL_ELEMENTS[idx] in ('Co','Ni','Mn')]
    ## set swap generators
    for g in ENGINE.groups:
        idx   = g.indexes[0]
        elIdx = ALL_ELEMENTS[idx]
        if elIdx in ('li','Li'):
            SPG=SwapPositionsGenerator(swapList=meSwaps)
            g.set_move_generator(SPG)
        elif elIdx in ('co','Co','mn','Mn','ni','Ni'):
            SPG=SwapPositionsGenerator(swapList=liSwaps)
            g.set_move_generator(SPG)
        else:
            g.set_move_generator( TranslationGenerator(amplitude=0.05) )
    ## run engine
    ENGINE.run(numberOfSteps=numberOfSteps, saveFrequency=saveFrequency)

def removes_run(numberOfSteps=100, saveFrequency=100):
    ## compute indexes lists
    ALL_ELEMENTS = ENGINE.get_original_data('allElements')
    oIndexes  = [idx for idx in range(len(ALL_ELEMENTS)) if ALL_ELEMENTS[idx]=='o'  or ALL_ELEMENTS[idx]=='O']
    liIndexes = [idx for idx in range(len(ALL_ELEMENTS)) if ALL_ELEMENTS[idx]=='li' or ALL_ELEMENTS[idx]=='Li']
    meIndexes = [idx for idx in range(len(ALL_ELEMENTS)) if ALL_ELEMENTS[idx] in ('co','ni','mn') or ALL_ELEMENTS[idx] in ('Co','Ni','Mn')]
    ## create empty group to remove oxygen.
    ## By default EmptyGroup move generator is AtomsRemoveGenerator with its atomsList
    ## None which means it will remove any atom from system.
    RO  = EmptyGroup()
    RO.moveGenerator.set_maximum_collected(20)
    RO.moveGenerator.set_atoms_list(oIndexes)
    ## create empty group to remove lithium
    RLi = EmptyGroup()
    RLi.moveGenerator.set_maximum_collected(500)
    RLi.moveGenerator.set_atoms_list(liIndexes)
    ## create empty group to remove cobalt
    RCo = EmptyGroup()
    RCo.moveGenerator.set_maximum_collected(500)
    RCo.moveGenerator.set_atoms_list(meIndexes)
    ## set groups to engine
    ENGINE.set_groups([RO,RLi,RCo])
    ## run engine
    ENGINE.run(numberOfSteps=numberOfSteps, saveFrequency=saveFrequency)


#  ####################################################################################  #
#  ################################## RUN SIMULATION ##################################  #
## run normal
normal_run(numberOfSteps=100000, saveFrequency=50000)
PDF_CONSTRAINT.set_adjust_scale_factor((10,0.7,1.3))
normal_run(numberOfSteps=250000, saveFrequency=50000)

## add first swapping frame
if not ENGINE.is_frame('swap_1'):
    ENGINE.add_frame('swap_1')
## use swap_1 frame and run swapping
ENGINE.set_used_frame('swap_1')
swaps_run(numberOfSteps=200000, saveFrequency=50000)
normal_run(numberOfSteps=250000, saveFrequency=50000)

## add first removing frame
if not ENGINE.is_frame('removes_2'):
    ENGINE.add_frame('removes_2')
## use removes_2 frame and run some removes
ENGINE.set_used_frame('removes_2')
for _ in range(20):
    # remove as little as possible then try to refine
    removes_run(numberOfSteps=100, saveFrequency=100)
    normal_run(numberOfSteps=100000, saveFrequency=10000)
    swaps_run(numberOfSteps=50000, saveFrequency=10000)





##########################################################################################
###################################### CALL plot.py ######################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, FLOAT_TYPE
from fullrmc.Engine import Engine
from fullrmc.Core.Collection import rebin
from fullrmc.Constraints.AtomicCoordinationConstraints import AtomicCoordinationNumberConstraint
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.StructureFactorConstraints import ReducedStructureFactorConstraint
from fullrmc.Generators.Swaps import SwapPositionsGenerator


##########################################################################################
#####################################  CREATE ENGINE  ####################################
# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
# files name
engineFileName = "engine.rmc"
grFileName     = "experimental.gr"
sqFileName     = "experimental.fq"
pdbFileName    = "system.pdb"
# engine variables
grExpPath      = os.path.join(DIR_PATH, grFileName)
sqExpPath      = os.path.join(DIR_PATH, sqFileName)
pdbPath        = os.path.join(DIR_PATH, pdbFileName)
engineFilePath = os.path.join(DIR_PATH, engineFileName)
# set some useful flags
FRESH_START = False

# check Engine exists, if not build it otherwise load it.
ENGINE = Engine(path=None)
if not ENGINE.is_engine(engineFilePath) or FRESH_START:
   # create engine
    ENGINE = Engine(path=engineFilePath, freshStart=True)
    ENGINE.set_pdb(pdbFileName)
    # add G(r) constraint
    PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=grExpPath, weighting="atomicNumber")
    ENGINE.add_constraints([PDF_CONSTRAINT])
    # Rebin S(Q) experimental data and build constraint
    Sq = np.transpose( rebin(np.loadtxt(sqExpPath) , bin=0.05) ).astype(FLOAT_TYPE)
    RSF_CONSTRAINT = ReducedStructureFactorConstraint(experimentalData=Sq, weighting="atomicNumber")
    ENGINE.add_constraints([RSF_CONSTRAINT])
    # add coordination number constraint and set to un-used
    ACN_CONSTRAINT = AtomicCoordinationNumberConstraint()
    ENGINE.add_constraints([ACN_CONSTRAINT])
    #ACN_CONSTRAINT.set_used(False)
    # add inter-molecular distance constraint
    EMD_CONSTRAINT = InterMolecularDistanceConstraint(defaultDistance=2.2, flexible=True)
    ENGINE.add_constraints([EMD_CONSTRAINT])
    # save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(engineFilePath)
    # unpack constraints before fitting in case tweaking is needed
    PDF_CONSTRAINT, RSF_CONSTRAINT, ACN_CONSTRAINT, EMD_CONSTRAINT = ENGINE.constraints

##########################################################################################
#####################################  DIFFERENT RUNS  ###################################
def run_normal(nsteps, saveFrequency, engineFilePath):
    ENGINE.set_groups_as_atoms()
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=saveFrequency)

def run_swap(nsteps, saveFrequency, engineFilePath):
    # activate coordination number. Play with definition and set_used to True
    ACN_CONSTRAINT.set_coordination_number_definition( [ ('ti','ti',2.5, 3.5, 4, 8),
                                                         ('ti','ni',2.2, 3.1, 6, 10),
                                                         ('ni','ni',2.5, 3.5, 4, 8),
                                                         ('ni','ti',2.2, 3.1, 6, 10) ] )
    ACN_CONSTRAINT.set_used(False)
    # reset groups
    ENGINE.set_groups_as_atoms()
    #### set swap generators Ni-->Ti and Ti-->Ni ###
    allElements = ENGINE.allElements
    niSwapList = [[idx] for idx in range(len(allElements)) if allElements[idx]=='ni']
    tiSwapList = [[idx] for idx in range(len(allElements)) if allElements[idx]=='ti']
    # create swap generator
    toNiSG = SwapPositionsGenerator(swapList=niSwapList)
    toTiSG = SwapPositionsGenerator(swapList=tiSwapList)
    # set swap generator to groups
    for g in ENGINE.groups:
        if allElements[g.indexes[0]]=='ni':
            g.set_move_generator(toTiSG)
        elif allElements[g.indexes[0]]=='ti':
            g.set_move_generator(toNiSG)
    # run
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=saveFrequency)

##########################################################################################
#####################################  RUN SIMULATION  ###################################
## run normal 10 times for 10000 step each time
for _ in range(10):
    run_normal(nsteps=10000, saveFrequency=10000, engineFilePath=engineFilePath)

### start fitting scale factors each 10 accepted moves
PDF_CONSTRAINT.set_adjust_scale_factor((10, 0.8, 1.2))
RSF_CONSTRAINT.set_adjust_scale_factor((10, 0.8, 1.2))

## run normal 100 times for 9000 step each time
for _ in range(50):
    run_normal(nsteps=10000, saveFrequency=10000, engineFilePath=engineFilePath)

## run swaping 100 times for 1000 step each time
#ACN_CONSTRAINT.set_used(False)
for _ in range(100):
    run_swap(nsteps=1000, saveFrequency=1000, engineFilePath=engineFilePath)

## run normal 100 times for 9000 step each time
for _ in range(10):
    run_normal(nsteps=10000, saveFrequency=10000, engineFilePath=engineFilePath)


##########################################################################################
###################################### CALL plot.py ######################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



# standard libraries imports
from __future__ import print_function
import os

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "thf_engine.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    PDF, EMD, B, BA, IA = ENGINE.constraints
    # plot all constraints
    PDF.plot(show=False)
    EMD.plot(show=False)
    B.plot(lineWidth=2, nbins=20,  split='element', show=False)
    BA.plot(lineWidth=2, nbins=20, split='element', show=False)
    IA.plot(lineWidth=2, nbins=20, split='element', show=True )
else:
    print(mes)



"""
This is a C compiled module to compute transformations from 
real to reciprocal space and vice versa.
"""
from libc.math cimport sin
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
#from cython.parallel import prange

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32


# declare constants
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 FLOAT32_ZERO    = 0.0
cdef C_FLOAT32 FLOAT32_ONE     = 1.0
cdef C_FLOAT32 FLOAT32_FOUR    = 4.0
cdef C_FLOAT32 FLOAT32_PI      = 3.1415927
cdef C_INT32   INT32_ONE       = 1
cdef C_INT32   INT32_ZERO      = 0


cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x)
    C_FLOAT32 ceil(C_FLOAT32 x)

    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)    
def gr_to_sq(np.ndarray[C_FLOAT32, ndim=1] distances not None,
             np.ndarray[C_FLOAT32, ndim=1] gr not None,
             np.ndarray[C_FLOAT32, ndim=1] qrange not None,
             C_FLOAT32                     rho):
    """
    Transform pair correlation function g(r) to static structure factor S(q).
            
    :Arguments:
       #. distances (float32 (n,) numpy.ndarray): The g(r) bins positions in real space.
       #. gr (float32 (n,) numpy.ndarray): The pair correlation function g(r) data.
       #. qrange (float32 (m,) numpy.ndarray): The S(q) bins positions in reciprocal space.
       #. rho (float32) [default=1]: The number density of the system. 
       
    :Returns:
       #. sq (float32 (m,) numpy.ndarray): The static structure factor S(q) data.
    """
    cdef C_INT32 qidx, ridx, qdim, rdim
    cdef C_FLOAT32 dr, dq, q,r, fact
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] sq = np.ones((len(qrange),), dtype=NUMPY_FLOAT32)
    
    dr   = distances[1]-distances[0]
    qdim = <C_INT32>qrange.shape[0]
    rdim = <C_INT32>distances.shape[0]
    fact = FLOAT32_FOUR * FLOAT32_PI * rho
    # s(q) = 1+4*pi*Rho*INTEGRAL[r * g(r) * sin(qr)/q * dr]
    for qidx from INT32_ZERO <= qidx < qdim:
        q = qrange[qidx]
        for ridx from INT32_ZERO <= ridx < rdim:
            r = distances[ridx]
            sq[qidx] += fact * ( dr*r*(np.sin(q*r)/q)*(gr[ridx]-1) )
            #sq[qidx] += dr*r*(np.sin(q*r)/q)*(gr[ridx]-1) 
    return sq#FLOAT32_ONE + FLOAT32_FOUR*FLOAT32_PI*rho*sq
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)    
def Gr_to_sq(np.ndarray[C_FLOAT32, ndim=1] distances not None,
             np.ndarray[C_FLOAT32, ndim=1] Gr not None,
             np.ndarray[C_FLOAT32, ndim=1] qrange not None):
    """
    Transform pair distribution function G(r) to static structure factor S(q).
            
    :Arguments:
       #. distances (float32 (n,) numpy.ndarray): The G(r) bins positions in real space.
       #. Gr (float32 (n,) numpy.ndarray): The pair correlation function Gr) data.
       #. qrange (float32 (m,) numpy.ndarray): The S(q) bins positions in reciprocal space.
       
    :Returns:
       #. sq (float32 (m,) numpy.ndarray): The static structure factor S(q) data.
    """
    cdef C_INT32 qidx, ridx, qdim, rdim
    cdef C_FLOAT32 dr, dq, q,r, fact
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] sq = np.ones((len(qrange),), dtype=NUMPY_FLOAT32)
    
    dr   = distances[1]-distances[0]
    qdim = <C_INT32>qrange.shape[0]
    rdim = <C_INT32>distances.shape[0]
    # s(q) = INTEGRAL[G(r) * sin(qr)/q * dr]
    for qidx from INT32_ZERO <= qidx < qdim:
        q = qrange[qidx]
        for ridx from INT32_ZERO <= ridx < rdim:
            r = distances[ridx]
            sq[qidx] +=  dr*(np.sin(q*r)/q)*(Gr[ridx]) 
    return sq
    
 
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def sq_to_Gr(np.ndarray[C_FLOAT32, ndim=1] qValues not None,
             np.ndarray[C_FLOAT32, ndim=1] rValues not None,
             np.ndarray[C_FLOAT32, ndim=1] sq not None,):
    """
    Transform static structure factor S(q) to pair distribution function G(r).
            
    :Arguments:
       #. qValues (float32 (m,) numpy.ndarray): The S(q) bins positions in reciprocal space.
       #. rValues (float32 (n,) numpy.ndarray): The G(r) bins positions in real space.
       #. sq (float32 (m,) numpy.ndarray): The static structure factor S(q) data.
       
    :Returns:
       #. Gr (float32 (n,) numpy.ndarray): The pair correlation function Gr) data.
       
    """
    cdef C_INT32 ridx, rdim
    cdef C_FLOAT32 dq
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] Gr = np.zeros((len(rValues),), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] sinqr_dq = np.zeros((len(rValues),), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] sq_1  = sq-1
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] qsq_1 = qValues*sq_1
    
    rdim = <C_INT32>rValues.shape[0]
    dq   = qValues[1]-qValues[0]
    for ridx from INT32_ZERO <= ridx < rdim:
        sinqr_dq = dq * sin(qValues*rValues[ridx])
        Gr[ridx] = (2./np.pi) * np.sum( qsq_1 * sinqr_dq )
    return Gr  



# standard libraries imports
from __future__ import print_function
import os

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "SiOx.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    PDF, EMD, ACN = ENGINE.constraints

    PDF.plot(frame='size_distribution', multiplot=True, intra=False, shapeFunc=False,
             xlabelParams=True, ylabelParams=True, legendParams=False,
             titleFormat = "@{frame} [${multiframeWeight:.3f}\\%]$ ($Std.Err.={standardError:.3f}$)",
             show=False)

    PDF.plot(frame='size_distribution', multiplot=False, intra=False, shapeFunc=False,
             xlabelParams=True, ylabelParams=True, legendParams={'loc':'lower right', 'fontsize':8, "frameon":False},
             titleFormat = "@{frame} [${multiframeWeight:.3f}\\%]$ ($Std.Err.={standardError:.3f}$)",
             show=False)

    PDF.plot_multiframe_weights('size_distribution')

else:
    print(mes)



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, maxint
from fullrmc.Engine import Engine
from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Generators.Rotations import RotationGenerator
from fullrmc.Generators.Translations import TranslationTowardsCenterGenerator


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
#####################################  CREATE ENGINE  ####################################
pdbPath = "system.pdb"
ENGINE = Engine(path=None)
ENGINE.set_pdb(pdbPath)
# add inter-molecular distance constraint
EMD_CONSTRAINT = InterMolecularDistanceConstraint(defaultDistance=1.75)
ENGINE.add_constraints([EMD_CONSTRAINT])

##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def move_towards():
    # set only one molecule group
    ENGINE.set_groups_as_molecules()
    secMolIdxs = ENGINE.groups[1].indexes
    ENGINE.set_groups( ENGINE.groups[0] )
    # set move generator
    for g in ENGINE.groups:
        t = TranslationTowardsCenterGenerator(center={'indexes': secMolIdxs},amplitude=0.15, angle=90)
        r = RotationGenerator(amplitude=10)
        mg = MoveGeneratorCollector(collection=[t,r], randomize=True, weights=[(0,1),(1,5)])
        g.set_move_generator(mg)
    # set runtime parameters
    nsteps = 1000
    xyzFrequency = 1
    # run engine
    xyzPath="trajectory.xyz"
    if os.path.isfile(xyzPath): os.remove(xyzPath)
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps, xyzFrequency=xyzFrequency, xyzPath=xyzPath, restartPdb=None)



##########################################################################################
#####################################  RUN SIMULATION  ###################################
move_towards()


##########################################################################################
##################################  VISUALIZE SIMULATION  ################################
ENGINE.set_pdb(pdbPath)
ENGINE.visualize( commands = ["trajectory.xyz"],
                  boxWidth=0, bgColor='white',
                  representationParams='CPK 1.0 0.2 50 50')



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, FLOAT_TYPE
from fullrmc.Engine import Engine
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.PairCorrelationConstraints import PairCorrelationConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint
from fullrmc.Constraints.ImproperAngleConstraints import ImproperAngleConstraint
from fullrmc.Core.Collection import convert_Gr_to_gr
from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
from fullrmc.Core.GroupSelector import RecursiveGroupSelector
from fullrmc.Selectors.RandomSelectors import RandomSelector
from fullrmc.Selectors.OrderedSelectors import DefinedOrderSelector
from fullrmc.Generators.Translations import TranslationGenerator, TranslationAlongSymmetryAxisGenerator
from fullrmc.Generators.Rotations import RotationGenerator, RotationAboutSymmetryAxisGenerator
from fullrmc.Generators.Agitations import DistanceAgitationGenerator, AngleAgitationGenerator


##########################################################################################
#####################################  CREATE ENGINE  ####################################
# dirname
try:
    DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
except:
    DIR_PATH = ''

    
# engine file names
engineFileName = "thf_engine.rmc"
expFileName    = "thf_pdf.exp"
pdbFileName    = "thf.pdb"
freshStart     = False

# engine variables
expPath        = os.path.join(DIR_PATH, expFileName)
pdbPath        = os.path.join(DIR_PATH, pdbFileName)
engineFilePath = os.path.join(DIR_PATH, engineFileName)
ENGINE = Engine(path=None)
if freshStart or not ENGINE.is_engine(engineFilePath):
    # create engine
    ENGINE = Engine(path=engineFilePath, freshStart=True)
    ENGINE.set_pdb(pdbFileName)
    ## create experimental constraints
    #PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=expPath, weighting="atomicNumber")
    _,_,_, gr = convert_Gr_to_gr(np.loadtxt(expPath), minIndex=[4,5,6])
    dataWeights = np.ones(gr.shape[0])
    dataWeights[:np.nonzero(gr[:,1]>0)[0][0]] = 0
    PDF_CONSTRAINT = PairCorrelationConstraint(experimentalData=gr.astype(FLOAT_TYPE), weighting="atomicNumber", dataWeights=dataWeights)
    # create and define molecular constraints
    EMD_CONSTRAINT = InterMolecularDistanceConstraint(defaultDistance=1.5)
    B_CONSTRAINT   = BondConstraint()
    BA_CONSTRAINT  = BondsAngleConstraint()
    IA_CONSTRAINT  = ImproperAngleConstraint()
    # add constraints to engine
    ENGINE.add_constraints([PDF_CONSTRAINT, EMD_CONSTRAINT, B_CONSTRAINT, BA_CONSTRAINT, IA_CONSTRAINT])
    # initialize constraints definitions
    B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"THF": [('O' ,'C1' , 1.29, 1.70),
                                                                      ('O' ,'C4' , 1.29, 1.70),
                                                                      ('C1','C2' , 1.29, 1.70),
                                                                      ('C2','C3' , 1.29, 1.70),
                                                                      ('C3','C4' , 1.29, 1.70),
                                                                      ('C1','H11', 0.58, 1.15),('C1','H12', 0.58, 1.15),
                                                                      ('C2','H21', 0.58, 1.15),('C2','H22', 0.58, 1.15),
                                                                      ('C3','H31', 0.58, 1.15),('C3','H32', 0.58, 1.15),
                                                                      ('C4','H41', 0.58, 1.15),('C4','H42', 0.58, 1.15)] })
    BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"THF": [ ('O'  ,'C1' ,'C4' , 95 , 135),
                                                                          ('C1' ,'O'  ,'C2' , 95 , 135),
                                                                          ('C4' ,'O'  ,'C3' , 95 , 135),
                                                                          ('C2' ,'C1' ,'C3' , 90 , 120),
                                                                          ('C3' ,'C2' ,'C4' , 90 , 120),
                                                                          # H-C-H angle
                                                                          ('C1' ,'H11','H12', 95 , 125),
                                                                          ('C2' ,'H21','H22', 95 , 125),
                                                                          ('C3' ,'H31','H32', 95 , 125),
                                                                          ('C4' ,'H41','H42', 95 , 125),
                                                                          # H-C-O angle
                                                                          ('C1' ,'H11','O'  , 100, 120),
                                                                          ('C1' ,'H12','O'  , 100, 120),
                                                                          ('C4' ,'H41','O'  , 100, 120),
                                                                          ('C4' ,'H42','O'  , 100, 120),
                                                                          # H-C-C
                                                                          ('C1' ,'H11','C2' , 80, 123),
                                                                          ('C1' ,'H12','C2' , 80, 123),
                                                                          ('C2' ,'H21','C1' , 80, 123),
                                                                          ('C2' ,'H21','C3' , 80, 123),
                                                                          ('C2' ,'H22','C1' , 80, 123),
                                                                          ('C2' ,'H22','C3' , 80, 123),
                                                                          ('C3' ,'H31','C2' , 80, 123),
                                                                          ('C3' ,'H31','C4' , 80, 123),
                                                                          ('C3' ,'H32','C2' , 80, 123),
                                                                          ('C3' ,'H32','C4' , 80, 123),
                                                                          ('C4' ,'H41','C3' , 80, 123),
                                                                          ('C4' ,'H42','C3' , 80, 123) ] })
    IA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"THF": [ ('C2','O','C1','C4', -15, 15),
                                                                          ('C3','O','C1','C4', -15, 15) ] })
    # save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(engineFilePath)
    PDF_CONSTRAINT, EMD_CONSTRAINT, B_CONSTRAINT, BA_CONSTRAINT, IA_CONSTRAINT = ENGINE.constraints


##########################################################################################
#####################################  DIFFERENT RUNS  ###################################
# ############ RUN C-H BONDS ############ #
def bonds_CH(ENGINE, rang=10, recur=10, refine=False, explore=True):
    groups = []
    for idx in range(0,ENGINE.pdb.numberOfAtoms, 13):
        groups.append( np.array([idx+1 ,idx+2 ], dtype=np.int32) ) # C1-H11
        groups.append( np.array([idx+1 ,idx+3 ], dtype=np.int32) ) # C1-H12
        groups.append( np.array([idx+4 ,idx+5 ], dtype=np.int32) ) # C2-H21
        groups.append( np.array([idx+4 ,idx+6 ], dtype=np.int32) ) # C2-H22
        groups.append( np.array([idx+7 ,idx+8 ], dtype=np.int32) ) # C3-H31
        groups.append( np.array([idx+7 ,idx+9 ], dtype=np.int32) ) # C3-H32
        groups.append( np.array([idx+10,idx+11], dtype=np.int32) ) # C4-H41
        groups.append( np.array([idx+10,idx+12], dtype=np.int32) ) # C4-H42
    ENGINE.set_groups(groups)
    [g.set_move_generator(DistanceAgitationGenerator(amplitude=0.2,agitate=(True,True))) for g in ENGINE.groups]
    # set selector
    if refine or explore:
        gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
        ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'bonds_CH' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN H-C-H ANGLES ############ #
def angles_HCH(ENGINE, rang=5, recur=10, refine=False, explore=True):
    groups = []
    for idx in range(0,ENGINE.pdb.numberOfAtoms, 13):
        groups.append( np.array([idx+1 ,idx+2, idx+3 ], dtype=np.int32) ) # H11-C1-H12
        groups.append( np.array([idx+4 ,idx+5, idx+6 ], dtype=np.int32) ) # H21-C2-H22
        groups.append( np.array([idx+7 ,idx+8, idx+9 ], dtype=np.int32) ) # H31-C3-H32
        groups.append( np.array([idx+10,idx+11,idx+12], dtype=np.int32) ) # H41-C4-H42
    ENGINE.set_groups(groups)
    [g.set_move_generator(AngleAgitationGenerator(amplitude=5)) for g in ENGINE.groups]
    # set selector
    if refine or explore:
        gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
        ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'angles_HCH' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN ATOMS ############ #
def atoms_type(ENGINE, type='C', rang=30, recur=20, refine=False, explore=True):
    allElements = ENGINE.allElements
    groups = []
    for idx, el in enumerate(allElements):
        if el == type:
            groups.append( [idx] )
    ENGINE.set_groups(groups)
    # set selector
    if refine or explore:
        gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
        ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'atoms' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN ATOMS ############ #
def atoms(ENGINE, rang=30, recur=20, refine=False, explore=True):
    ENGINE.set_groups_as_atoms()
    # set selector
    if refine or explore:
        gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
        ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'atoms' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN ROTATION ABOUT SYMM AXIS 0 ############ #
def about0(ENGINE, rang=5, recur=100, refine=True, explore=False):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(RotationAboutSymmetryAxisGenerator(axis=0, amplitude=180)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'about0' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN ROTATION ABOUT SYMM AXIS 1 ############ #
def about1(ENGINE, rang=5, recur=10, refine=True, explore=False):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(RotationAboutSymmetryAxisGenerator(axis=1, amplitude=180)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'about1' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN ROTATION ABOUT SYMM AXIS 2 ############ #
def about2(ENGINE, rang=5, recur=100, refine=True, explore=False):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(RotationAboutSymmetryAxisGenerator(axis=2, amplitude=180)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'about2' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN TRANSLATION ALONG SYMM AXIS 0 ############ #
def along0(ENGINE, rang=5, recur=100, refine=False, explore=True):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(TranslationAlongSymmetryAxisGenerator(axis=0, amplitude=0.1)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'along0' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN TRANSLATION ALONG SYMM AXIS 1 ############ #
def along1(ENGINE, rang=5, recur=100, refine=False, explore=True):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(TranslationAlongSymmetryAxisGenerator(axis=1, amplitude=0.1)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    recur = 200
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'along1' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN TRANSLATION ALONG SYMM AXIS 2 ############ #
def along2(ENGINE, rang=5, recur=100, refine=False, explore=True):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(TranslationAlongSymmetryAxisGenerator(axis=2, amplitude=0.1)) for g in ENGINE.groups]
    # set selector
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(rang):
        LOGGER.info("Running 'along2' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ RUN MOLECULES ############ #
def molecules(ENGINE, rang=5, recur=100, refine=False, explore=True):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator( MoveGeneratorCollector(collection=[TranslationGenerator(amplitude=0.2),RotationGenerator(amplitude=2)],randomize=True) ) for g in ENGINE.groups]
    # number of steps
    nsteps = 20*len(ENGINE.groups)
    # set selector
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    for stepIdx in range(rang):
        LOGGER.info("Running 'molecules' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)

# ############ SHRINK SYSTEM ############ #
def shrink(ENGINE, newDim):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator( MoveGeneratorCollector(collection=[TranslationGenerator(amplitude=0.2),RotationGenerator(amplitude=5)],randomize=True) ) for g in ENGINE.groups]
    # get groups order
    centers   = [np.sum(ENGINE.realCoordinates[g.indexes], axis=0)/len(g) for g in ENGINE.groups]
    distances = [np.sqrt(np.add.reduce(c**2)) for c in centers]
    order     = np.argsort(distances)
    # change boundary conditions
    bcFrom = str([list(bc) for bc in ENGINE.boundaryConditions.get_vectors()] )
    ENGINE.set_boundary_conditions(newDim)
    bcTo   = str([list(bc) for bc in ENGINE.boundaryConditions.get_vectors()] )
    LOGGER.info("boundary conditions changed from %s to %s"%(bcFrom,bcTo))
    # set selector
    recur = 200
    gs = RecursiveGroupSelector(DefinedOrderSelector(ENGINE, order = order ), recur=recur, refine=True)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    for stepIdx in range(10):
        LOGGER.info("Running 'shrink' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps)
        fname = "shrink_"+str(newDim).replace(".","p")

##########################################################################################
#####################################  RUN SIMULATION  ###################################
atoms(ENGINE, explore=True, refine=False)
atoms(ENGINE, explore=True, refine=False)
# set short limits
PDF_CONSTRAINT.set_limits((None,5))
# fit bonds
bonds_CH(ENGINE)
# fit angles
angles_HCH(ENGINE)
# reset limits to normal
PDF_CONSTRAINT.set_limits((None,None))
# fit atoms
atoms(ENGINE, explore=False, refine=False)
atoms_type(ENGINE, explore=False, refine=False)
# fit molecules
molecules(ENGINE)

# refine scaling factor
atoms(ENGINE, explore=True, refine=False)
atoms_type(ENGINE, explore=True, refine=False)
PDF_CONSTRAINT.set_limits((None,None))
atoms(ENGINE, explore=False, refine=False)
about0(ENGINE)
about1(ENGINE)
about2(ENGINE)
along0(ENGINE)
along1(ENGINE)
along2(ENGINE)
molecules(ENGINE)
atoms(ENGINE, explore=True, refine=False)
# allow scale adjustment
PDF_CONSTRAINT.set_adjust_scale_factor((10, 0.8, 1.2))
about0(ENGINE)
about1(ENGINE)
about2(ENGINE)
along0(ENGINE)
along1(ENGINE)
along2(ENGINE)
molecules(ENGINE)
atoms(ENGINE, explore=True, refine=False)


##########################################################################################
###################################### CALL plot.py ######################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



"""
.. py:function:: get_examples_path()

    Get distribution examples directory path in the standard distribution.

    :Returns:
        #. path (string): directory path in a form of a string.


.. py:function:: find_example(name)

    Return example's path given its name.

    :Parameters:
        #. name (string): The name of the example.

    :Returns:
        #. path (None, string): The path of the example if found.
           None if example is not found.


.. py:function:: get_available_examples()

    Get all available examples as delivered in the standard distribution.

    :Returns:
        #. examples (list): List of examples paths.


.. py:function:: print_available_examples()

    Print all available examples path in the standard distribution.
"""

# standard distribution imports
from __future__ import print_function
import os

def get_examples_path():
    path, _ = os.path.split(os.path.abspath(__file__))
    return path

def find_example(name):
    if not isinstance(name, basestring):
        return None
    name = str(name)
    if not len(name):
        return None
    path = get_examples_path()
    exPath = os.path.join(path, name)
    if os.path.isdir(exPath):
        return exPath
    else:
        return None

def get_available_examples():
    path = get_examples_path()
    # create examples list
    examples = []
    for d in os.listdir( path ):
        d = os.path.join(path, d)
        if os.path.isdir(d):
            examples.append(d)
    return examples

def print_available_examples():
    for e in get_available_examples():
        print(e)



"""
This module provides all the global types, variables and some general methods that fullrmc needs.

Types:
======
#. INT_TYPE: The integer data type adopted by fullrmc.
#. FLOAT_TYPE: The floating data type adopted by fullrmc.

Variables:
==========
#. PRECISION: The floating data type precision variable.
#. FLOAT_PLUS_INFINITY: The floating data type maximum possible number.
#. FLOAT_MINUS_INFINITY: The floating data type minimum possible number.
#. INT_PLUS_INFINITY: The integer data type maximum possible number.
#. INT_MINUS_INFINITY: The integer data type minimum possible number.
#. PI: pi the ratio of a circle's circumference to its diameter.
"""
# standard libraries imports
from __future__ import print_function
import sys
import os

# external libraries imports
import numpy as np
from pysimplelog import SingleLogger as LOG


# data types definitions
INT_TYPE   = np.int32   # must be the integer type for the whole package
FLOAT_TYPE = np.float32 # must be the float type for the whole package

# floating precision
if FLOAT_TYPE is np.float32:
    PRECISION = FLOAT_TYPE(1e-5)  # Used to check precision of float32 loaded data
elif FLOAT_TYPE is np.float64:
    PRECISION = FLOAT_TYPE(1e-10) # Used to check precision of float64 loaded data
else:
    raise Exception("Unknown float type '%s'"%FLOAT_TYPE)

# Constants definitions
FLOAT_PLUS_INFINITY  = FLOAT_TYPE(np.finfo(FLOAT_TYPE).max) # +inf number for float type
FLOAT_MINUS_INFINITY = FLOAT_TYPE(np.finfo(FLOAT_TYPE).min) # -inf number for float type
INT_PLUS_INFINITY    = INT_TYPE(np.iinfo(np.int32).max)     # +inf number for integer type
INT_MINUS_INFINITY   = INT_TYPE(np.iinfo(np.int32).min)     # -inf number for integer type
PI                   = FLOAT_TYPE(np.pi)                    # pi the ratio of a circle's circumference to its diameter, set as constant for typing
WATER_NUMBER_DENSITY = FLOAT_TYPE(.0333679)                 # This value is in A^-3

## get python 2 and 3 compatibilities
IS_PY3 = int(sys.version_info.major)==3
if IS_PY3:
    # THIS IS PYTHON 3
    str        = str
    long       = int
    unicode    = str
    bytes      = bytes
    basestring = str
    xrange     = range
    range      = lambda *args: list( xrange(*args) )
    maxint     = sys.maxsize
else:
    # THIS IS PYTHON 2
    assert int(sys.version_info.major)==2, "Only python 2.x.y or python 3.x.y are allowed. %s is used instead"%(sys.version.split()[0],)
    assert int(sys.version_info.minor)>=7, "Only minor version 7 is allowed in python 2.x.y distribution. %s is used instead"%(sys.version.split()[0],)
    str        = str
    unicode    = unicode
    bytes      = str
    long       = long
    basestring = basestring
    xrange     = xrange
    range      = range
    maxint     = sys.maxint


# Create LOGGER
class Logger(LOG):
    def custom_init(self):
        # set logfile basename
        logFile = os.path.join(os.getcwd(), "fullrmc")
        self.set_log_file_basename(logFile)
        # set new log types
        self.add_log_type("move not tried", name="INFO",           level=-15)
        self.add_log_type("move rejected",  name="INFO",           level=-10)
        self.add_log_type("move accepted",  name="INFO",           level= 15)
        self.add_log_type("engine saved",   name="INFO",           level= 17)
        self.add_log_type("argument fixed", name="FIXED",          level= 20)
        self.add_log_type("implement",      name="IMPLEMENTATION", level= 100)
        self.add_log_type("usage",          name="USAGE",          level= 1000)
        self.add_log_type("frame",          name="FRAME",          level= 1000)
        self.add_log_type("report",         name="REPORT ISSUE",   level= 1000)
        # set minimum level to 10
        self.set_minimum_level(10)
        # force error and critical logging no matter what logging level is
        self.force_log_type_flags(logType="error",    stdoutFlag=True, fileFlag=True)
        self.force_log_type_flags(logType="critical", stdoutFlag=True, fileFlag=True)
        #self.force_log_type_flags(logType="implement",stdoutFlag=True, fileFlag=True)
        #self.force_log_type_flags(logType="usage",    stdoutFlag=True, fileFlag=True)

    def fixed(self, message):
        """alias to message at fixed level"""
        self.log("argument fixed", message)

    def accepted(self, message):
        """alias to message at move accepted level"""
        self.log("move accepted", message)

    def rejected(self, message):
        """alias to message at move rejected level"""
        self.log("move rejected", message)

    def nottried(self, message):
        """alias to message at move not tried level"""
        self.log("move not tried", message)

    def saved(self, message):
        """alias to message at save engine level"""
        self.log("engine saved", message)

    def impl(self, message):
        """alias to message at implement engine level"""
        self.log("implement", message)

    def implement(self, message):
        """alias to message at usage engine level"""
        self.log("usage", message)

    def usage(self, message):
        """alias to message at usage engine level"""
        self.log("usage", message)

    def frame(self, message):
        """alias to message at setting engine level"""
        self.log("frame", message)

    def report(self, message):
        """alias to message at report engine level"""
        self.log("report", message)

# initialize Logger
LOGGER = Logger(name="fullrmc")
